/* print.c - print game and player lists.
 *
 * Copyright (C) 2001  Jochen Voss.  */

static const  char  rcsid[] = "$Id: print.c,v 1.2 2001/09/01 16:40:55 voss Exp $";

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <iconv.h>

#include <gnome.h>
#include <libgnomeprint/gnome-print.h>
#include <libgnomeprint/gnome-print-master.h>
#include <libgnomeprint/gnome-print-master-preview.h>

#include <go-player.h>
#include <go-game.h>
#include <go-client.h>
#include <go-strengths.h>
#include <go.h>

#include "golist.h"


#define  PAPER  "a4"
#define  TOP_MARGIN  40
#define  BOTTOM_MARGIN  60
#define  LEFT_MARGIN  50


/**********************************************************************
 * Print the game list into a GnomePrintContext.
 */

struct print_state {
  GnomePrintContext *pc;
  const GnomePaper *paper;
  GnomeFont *rm_font, *bf_font;

  gboolean  first;
  double  pos, xoffset;
  time_t  last_date;
  int  page;
};

static double
print_pos_top (const struct print_state *ps)
{
  return   gnome_paper_psheight (ps->paper) - TOP_MARGIN;
}

static void
print_aligned (struct print_state *ps, double xmin, double xmax, int align,
	       GnomeFont *font, const char *str)
{
  char  buffer [1024], *ptr, *strb, *strp;
  iconv_t cd;
  double  w, mag;
  int  res;
  size_t  x, in, out;

  mag = LEFT_MARGIN + ps->xoffset;
  
  strb = strp = g_strdup (str);
  cd = iconv_open ("utf8", "latin1");
  if (cd == (iconv_t)-1) {
    perror ("iconv_open");
    abort ();
  }

  in = strlen(strp);
  out = 1024;
  ptr = buffer;
  x = iconv (cd, &strp, &in, &ptr, &out);
  *ptr = '\0';

  res = iconv_close (cd);
  if (res) {
    perror ("iconv_close");
    abort ();
  }
  g_free (strb);
  
  w = gnome_font_get_width_string (font, buffer);
  gnome_print_setfont (ps->pc, font);
  if (align > 0) {
    gnome_print_moveto (ps->pc, mag + xmax-w, ps->pos);
  } else if (align < 0) {
    gnome_print_moveto (ps->pc, mag + xmin, ps->pos);
  } else {
    gnome_print_moveto (ps->pc, mag + xmin+0.5*(xmax-xmin-w), ps->pos);
  }
  gnome_print_show (ps->pc, buffer);
}

/**********************************************************************
 * Print the game list into a GnomePrintContext.
 */

static void
print_header (struct print_state *ps)
{
  ps->pos -= 15;
  print_aligned (ps, 70, 140, -1, ps->bf_font, "black");
  print_aligned (ps, 150, 220, -1, ps->bf_font, "white");
  print_aligned (ps, 230, 290, 0, ps->bf_font, "board");
  print_aligned (ps, 300, 350, 1, ps->bf_font, "hcap");
  print_aligned (ps, 350, 400, 1, ps->bf_font, "komi");
  ps->pos -= 10;
}  

static void
print_one_game (GoClient *c, GoGame *g, int i, int n, void *data)
{
  struct print_state *ps = data;
  char  buffer [64];
  gboolean  show_date;

  if (g->date
      && (ps->last_date != g->date)) {
    if (! ps->first)  ps->pos -= 5;
    show_date = TRUE;
    ps->last_date = g->date;
  } else {
    show_date = FALSE;
  }
  ps->first = FALSE;
  if (ps->pos <= BOTTOM_MARGIN) {
    char  buffer [64];

    ps->pos = BOTTOM_MARGIN - 30;
    snprintf (buffer, 64, "- %d -", ps->page);
    print_aligned (ps, 0, gnome_paper_pswidth (ps->paper)-2*LEFT_MARGIN, 0,
		   ps->rm_font, buffer);
    gnome_print_showpage (ps->pc);

    ++ ps->page;

    snprintf (buffer, 64, "%d", ps->page);
    gnome_print_beginpage (ps->pc, buffer);
    ps->pos = print_pos_top (ps);
    print_header (ps);
  }

  ps->pos -= 15;
  if (show_date) {
    struct tm *loctime = localtime (&((GoGame *)g)->date);
    strftime (buffer, 64, "%x", loctime);
    
    print_aligned (ps, 0, 60, -1, ps->rm_font, buffer);
  } else if (g->date) {
    print_aligned (ps, 5, 60, -1, ps->rm_font, "~");
  }

  print_aligned (ps, 70, 140, -1,
		 g->res == res_BLACK ? ps->bf_font : ps->rm_font,
		 GO_PLAYER(g->black)->token);
  print_aligned (ps, 150, 220, -1,
		 g->res == res_WHITE ? ps->bf_font : ps->rm_font,
		 GO_PLAYER(g->white)->token);

  print_aligned (ps, 230, 290, 0, ps->rm_font, board_to_str (g->board));
  g_snprintf (buffer, 64, "%d", g->handicap);
  print_aligned (ps, 300, 340, 1, ps->rm_font, buffer);
  if (g->with_jigo) {
    g_snprintf (buffer, 64, "%d", g->komi);
  } else {
    g_snprintf (buffer, 64, "%g", g->komi+0.5);
  }
  print_aligned (ps, 350, 390, 1, ps->rm_font, buffer);
}

static void
print_all_games (GoClient *client, GnomePrintContext *pc)
{
  char  buffer [64];
  struct print_state  ps;

  ps.pc = pc;
  ps.paper = gnome_paper_with_name (PAPER);
  ps.rm_font = gnome_font_new_closest ("Times", GNOME_FONT_BOOK, FALSE, 12);
  g_assert (ps.rm_font);
  ps.bf_font = gnome_font_new_closest ("Times", GNOME_FONT_BOLD, FALSE, 12);
  g_assert (ps.bf_font);
  
  ps.page = 1;

  gnome_print_beginpage (ps.pc, "1");
  ps.first = TRUE;
  ps.last_date = 0;
  ps.pos = print_pos_top (&ps);
  ps.xoffset = 0;
  print_header (&ps);
  go_client_foreach_game (client, print_one_game, &ps);

  ps.pos = BOTTOM_MARGIN - 30;
  snprintf (buffer, 64, "- %d -", ps.page);
  print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, 0,
		 ps.rm_font, buffer);
  gnome_print_showpage (ps.pc);
}

void
print_games (GtkObject *client)
{
  GnomePrintMaster *gpm;
  GnomePrintContext *gpc;
  GnomePrintMasterPreview *pmp;

  gpm = gnome_print_master_new ();
  gnome_print_master_set_paper (gpm, gnome_paper_with_name (PAPER));
  gpc = gnome_print_master_get_context (gpm);
  print_all_games (GO_CLIENT (client), GNOME_PRINT_CONTEXT (gpc));
  gnome_print_master_close (gpm);

  pmp = gnome_print_master_preview_new_with_orientation
    (gpm, _("Print preview"), FALSE);
  gtk_widget_show (GTK_WIDGET (pmp));

  gtk_object_unref (GTK_OBJECT (gpm));
}

/**********************************************************************
 * Print the list of players into a GnomePrintContext.
 */

static void
print_one_player (GoClient *c, GoPlayer *p, int i, int n, void *data)
{
  struct print_state *ps = data;
  
  ps->pos -= 15;
    
  print_aligned (ps, 0, 140, -1, ps->rm_font, p->name);
  print_aligned (ps, 150, 200, -1, ps->rm_font, p->token);
  if (go_data_player_component (c->data, p) == 0 && p->strength > -100) {
    char  buffer [64];
    
    snprintf (buffer, 64, "%.1f", p->strength);
    print_aligned (ps, 210, 250, +1, ps->rm_font, buffer);
    snprint_strength (buffer, 64, p->strength);
    print_aligned (ps, 260, 300, +1, ps->rm_font, buffer);
  }
  if (p->rank) {
    print_aligned (ps, 300, 310, 0, ps->bf_font, "*");
  }
}

static void
print_all_players (GoClient *client, GnomePrintContext *pc)
{
  struct print_state  ps;

  ps.pc = pc;
  ps.paper = gnome_paper_with_name (PAPER);
  ps.rm_font = gnome_font_new_closest ("Times", GNOME_FONT_BOOK, FALSE, 12);
  g_assert (ps.rm_font);
  ps.bf_font = gnome_font_new_closest ("Times", GNOME_FONT_BOLD, FALSE, 12);
  g_assert (ps.bf_font);
  
  ps.page = 1;

  gnome_print_beginpage (ps.pc, "players");
  ps.pos = print_pos_top (&ps);
  ps.xoffset = 0;

  ps.pos -= 15;
  print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		 ps.bf_font, "Spielerliste fr den geheimen Go-Spielabend");
  ps.pos -= 15;
  {
    time_t curtime;
    struct tm *loctime;
    char  buffer [256];

    curtime = time (NULL);
    loctime = localtime (&curtime);
    strftime (buffer, 256, "Stand: %x", loctime);
    
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.bf_font, buffer);
  }

  ps.pos -= 15;
  go_client_foreach_player (client, print_one_player, &ps);
  gnome_print_showpage (ps.pc);
}

void
print_players (GtkObject *client)
{
  GnomePrintMaster *gpm;
  GnomePrintContext *gpc;
  GnomePrintMasterPreview *pmp;

  gpm = gnome_print_master_new ();
  gnome_print_master_set_paper (gpm, gnome_paper_with_name (PAPER));
  gpc = gnome_print_master_get_context (gpm);
  print_all_players (GO_CLIENT (client), GNOME_PRINT_CONTEXT (gpc));
  gnome_print_master_close (gpm);

  pmp = gnome_print_master_preview_new_with_orientation
    (gpm, _("Print preview"), FALSE);
  gtk_widget_show (GTK_WIDGET (pmp));

  gtk_object_unref (GTK_OBJECT (gpm));
}

static GoGame *
last_player_game (GoClient *client, GoPlayer *player)
{
  int  i;

  for (i=go_client_game_count (client)-1; i>=0; --i) {
    GoGame *g = go_client_game (client, i);
    if (g->black == player || g->white == player)  return g;
  }
  return  NULL;
}

static void
last_pair_games (GoClient *client, GoPlayer *a, GoPlayer *b,
		 GoGame **res1, GoGame **res2)
{
  int  i;
  gboolean  second = FALSE;
  
  *res1 = *res2 = NULL;
  for (i=go_client_game_count(client)-1; i>=0; --i) {
    GoGame *g = go_client_game (client, i);
    if ((g->black == a && g->white == b)
	|| (g->black == b && g->white == a)) {
      if (second) {
	*res2 = g;
	return;
      }
      *res1 = g;
      second = TRUE;
    }
  }
}

static void
print_game_short (struct print_state *ps, GoGame *g)
{
  gchar *tmpl, *td, *str;

  ps->pos -= 12;
  print_aligned (ps, 12, 62, -1,
		 g->res == res_BLACK ? ps->bf_font : ps->rm_font,
		 GO_PLAYER(g->black)->token);
  print_aligned (ps, 70, 120, -1,
		 g->res == res_WHITE ? ps->bf_font : ps->rm_font,
		 GO_PLAYER(g->white)->token);
  if (g->date) {
    char  buffer [64];
    struct tm *loctime = localtime (&((GoGame *)g)->date);
    strftime (buffer, 64, "%x", loctime);
    td = g_strdup_printf (" (%s)", buffer);
  } else {
    td = NULL;
  }
  tmpl = g_strconcat ("%dx%d %dh %gk", td, NULL);
  g_free (td);

  str = g_strdup_printf (tmpl,
			 g->board, g->board, g->handicap,
			 g->with_jigo ? g->komi : g->komi+0.5);
  print_aligned (ps, 130, gnome_paper_pswidth (ps->paper)-2*LEFT_MARGIN, -1,
		 ps->rm_font, str);
  g_free (str);
  g_free (tmpl);
}

static void
suggest (struct print_state *ps, GoData *data, GoPlayer *a, GoPlayer *b,
	 GoStrengths *strengths, GoBoardSize board)
{
  int  hc;
  double  komi;
  gchar *stra, *strb, *str;

  if (a->strength > b->strength) {
    GoPlayer *c;
    c = a;  a = b;  b = c;
  }
  hc = go_strengths_find_handicap (strengths, a, b, board);
  if (board > 9 && hc > 9)  return;
  komi = go_strengths_find_komi (strengths, a, b, board, hc);
  ps->pos -= 12;
  print_aligned (ps, 12, 62, -1, ps->rm_font, GO_PLAYER(a)->token);
  print_aligned (ps, 70, 120, -1, ps->rm_font, GO_PLAYER(b)->token);
  stra = g_strdup_printf ("%dx%d %dh %gk", board, board, hc, komi);

  if (hc < 9) {
    hc = hc ? hc+1 : 2;
    komi = go_strengths_find_komi (strengths, a, b, board, hc);
    strb = g_strdup_printf (" or %dh %gk", hc, komi);
  } else {
    strb = NULL;
  }

  str = g_strconcat (stra, strb, NULL);
  print_aligned (ps, 130, gnome_paper_pswidth (ps->paper)-2*LEFT_MARGIN, -1,
		 ps->rm_font, str);
  g_free (str);
  g_free (stra);
  g_free (strb);
}

static void
do_print_sheet (GoClient *client, GoPlayer *p, GnomePrintContext *pc)
{
  struct print_state  ps;
  char  buffer [256];
  int  i;

  ps.pc = pc;
  ps.paper = gnome_paper_with_name (PAPER);
  ps.rm_font = gnome_font_new_closest ("Times", GNOME_FONT_BOOK, FALSE, 10);
  g_assert (ps.rm_font);
  ps.bf_font = gnome_font_new_closest ("Times", GNOME_FONT_BOLD, FALSE, 10);
  g_assert (ps.bf_font);
  
  ps.page = 1;

  gnome_print_beginpage (ps.pc, "players");
  ps.pos = print_pos_top (&ps);
  ps.xoffset = 0;

  ps.pos -= 12;
  g_snprintf (buffer, 256, _("player sheet for %s"), p->name);
  print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		 ps.bf_font,
		 buffer);
  ps.pos -= 12;
  {
    time_t curtime;
    struct tm *loctime;

    curtime = time (NULL);
    loctime = localtime (&curtime);
    strftime (buffer, 256, _("date: %x"), loctime);
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.bf_font, buffer);
  }
  
  ps.pos -= 24;
  g_snprintf (buffer, 256, _("token: %s"), p->token);
  print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		 ps.rm_font, buffer);
  if (p->rank) {
    g_snprintf (buffer, 256, _("nominal rank: %s"), p->rank);
    ps.pos -= 12;
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.rm_font, buffer);
  } else if (go_data_player_component (client->data, p) == 0
	     && p->strength > -100) {
    char  s [32];
    snprint_strength (s, 32, p->strength);
    g_snprintf (buffer, 256, _("estimated rank: %s"), s);
    ps.pos -= 12;
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.rm_font, buffer);
  }
  {
    int  n, k;

    n = go_data_player_game_count (client->data, p);
    k = go_data_player_win_count (client->data, p);
    g_snprintf (buffer, 256, _("number of games: %d"), n);
    ps.pos -= 12;
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.rm_font, buffer);
    g_snprintf (buffer, 256, _("number of wins: %d (%.1f%%)"), k, k*100.0/n);
    ps.pos -= 12;
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.rm_font, buffer);
  }
  {
    GoGame *g = last_player_game (client, p);
    if (g && g->date) {
      struct tm *loctime = localtime (&g->date);
      strftime (buffer, 256, _("last recorded game: %x"), loctime);
      ps.pos -= 12;
      print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		     ps.rm_font, buffer);
    }
  }
  ps.pos -= 12;

  for (i=0; i<go_client_player_count (client); ++i) {
    GoPlayer *opponent = go_client_player (client, i);
    GoGame *g1, *g2;

    if (p == opponent)  continue;
    last_pair_games (GO_CLIENT (client), p, opponent, &g1, &g2);
    if (! (g1 || g2 || go_data_players_equiv (client->data, p, opponent)))
      continue;

    if (go_data_player_component (client->data, opponent) == 0
	&& opponent->strength > -100) {
      char  s [32];
      snprint_strength (s, 32, opponent->strength);
      g_snprintf (buffer, 256, "%s (%s):", opponent->name, s);
    } else {
      g_snprintf (buffer, 256, "%s:", opponent->name);
    }
    if (ps.pos <= BOTTOM_MARGIN + 6*12) {
      ps.pos = print_pos_top (&ps);
      ps.xoffset = 0.5*gnome_paper_pswidth (ps.paper)-LEFT_MARGIN+10;
    }
    ps.pos -= 12;
    print_aligned (&ps, 0, gnome_paper_pswidth (ps.paper)-2*LEFT_MARGIN, -1,
		   ps.rm_font, buffer);
    if (g2)  print_game_short (&ps, g2);
    if (g1)  print_game_short (&ps, g1);
    if (go_data_players_equiv (client->data, p, opponent)) {
      suggest (&ps, client->data, p, opponent, GO_STRENGTHS(strengths), b19);
      suggest (&ps, client->data, p, opponent, GO_STRENGTHS(strengths), b13);
      suggest (&ps, client->data, p, opponent, GO_STRENGTHS(strengths), b9);
    }
    ps.pos -= 12;
  }
  gnome_print_showpage (ps.pc);
}

void
print_sheet (GtkObject *client, GoPlayer *player)
{
  GnomePrintMaster *gpm;
  GnomePrintContext *gpc;
  GnomePrintMasterPreview *pmp;

  gpm = gnome_print_master_new ();
  gnome_print_master_set_paper (gpm, gnome_paper_with_name (PAPER));
  gpc = gnome_print_master_get_context (gpm);
  do_print_sheet (GO_CLIENT (client), player, GNOME_PRINT_CONTEXT (gpc));
  gnome_print_master_close (gpm);

  pmp = gnome_print_master_preview_new_with_orientation
    (gpm, _("Print preview"), FALSE);
  gtk_widget_show (GTK_WIDGET (pmp));

  gtk_object_unref (GTK_OBJECT (gpm));
}
